/* NetPipe.c v 1.0 by Aaron Conole aka orgcandman
 *  tested and run on Linux v. 2.4.20, glibc v. 2.2.5
 *  platforms: x86
 * 
 *  This version features:
 *  -  TCP socket redirection
 *  
 */

/* handy dandy includes */

#include <stdlib.h>
#include <malloc.h>
#include <sys/select.h>
#include <sys/types.h>

#include <sys/time.h>		/* timeval, time_t */
#include <setjmp.h>		/* jmp_buf et al */
#include <sys/socket.h>		/* basics, SO_ and AF_ defs, sockaddr, ... */
#include <netinet/in.h>		/* sockaddr_in, htons, in_addr */
#include <netinet/in_systm.h>	/* misc crud that netinet/ip.h references */
#include <netinet/ip.h>		/* IPOPT_LSRR, header stuff */
#include <netinet/tcp.h>
#include <netdb.h>		/* hostent, gethostby*, getservby* */
#include <arpa/inet.h>		/* inet_ntoa */
#include <arpa/nameser.h>
#include <resolv.h>
#include <stdio.h>
#include <string.h>		/* strcpy, strchr, yadda yadda */
#include <errno.h>
#include <signal.h>
#include <fcntl.h>		/* O_WRONLY et al */
#include <getopt.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <net/if.h>

/* begin shameless theft of cheops code */
#include <linux/sockios.h>	/* GLIBC don't have sockios.h? */

typedef struct {
  char ifname[17];
  struct in_addr addr;
} interfacerec;

typedef struct {
  struct in_addr addr;
  unsigned long naddr;	/* netmask */
  interfacerec *iface;
} routerec;

static short numinterfaces, numroutes;
static interfacerec *interfaces;
static routerec *routes;

/* Standard Macro */
#ifndef MIN
#define MIN(x,y) (x<y) ? x : y;
#endif


struct in_addr getlocalip (unsigned int dest) {
  static struct in_addr ina;
  int i;
  
  for (i = 0; i < numroutes; i++){
    if ((dest & routes[i].naddr) == (unsigned long) routes[i].addr.s_addr){
      return (routes[i].iface->addr);
    }
  }
  
  ina.s_addr = 0;
  return ina;
}

void init_route_tables(void){
  int ifsock, i, i1, found;
  struct ifconf ifc;
  struct ifreq *ifr;
  char buf[1024], iface[16], *ptr;
  FILE *f;

  /* Create a channel to the NET kernel. */
  if ((ifsock = socket (AF_INET, SOCK_DGRAM, 0)) < 0){
    perror ("socket");
    exit (EXIT_FAILURE);
  }
  
  ifc.ifc_len = sizeof (buf);
  ifc.ifc_buf = buf;
  if (ioctl (ifsock, SIOCGIFCONF, &ifc) < 0){
    perror ("opening interface socket");
    close (ifsock);
    exit (EXIT_FAILURE);
  }
  
  /* Mental note: the ioctl fills ifc with all the interfaces on		*/
  /* the system.  The length divided by the size of the ifreq			*/
  /* struct is how many interfaces we have.					*/
  numinterfaces = (ifc.ifc_len / sizeof (struct ifreq));
  
  /* Allocate an array of interfacerec strutures to hold interface	        */
  /* configuration information.							*/
  
  interfaces = (interfacerec *) malloc (numinterfaces * sizeof (interfacerec));

  ifr = ifc.ifc_req;
  for (i = 0; i < numinterfaces; i++, ifr++){
    strcpy (interfaces[i].ifname, ifr->ifr_name);
    memcpy (&interfaces[i].addr, &((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr, sizeof (struct in_addr));
				/* What does ioctl return on systems that allow unnumbered		*/
				/* interface?																									*/
    if (ioctl (ifsock, SIOCGIFADDR, ifr) < 0)
      printf ("Couldn't get address for %s\n", ifr->ifr_name);
  }
  close (ifsock);
  
  if ((f = fopen ("/proc/net/route", "r")) == NULL){
    perror ("opening /proc/net/route");
    exit (EXIT_FAILURE);
  }
  
  /* Count the number of lines in /proc/net/route to determine the	*/
  /* size of the routing tables.																		*/
  numroutes = 0;
  fgets (buf, sizeof (buf), f);         /* strip out description line */
  while (!feof (f)){
    fgets (buf, sizeof (buf), f);
    numroutes++;
  }
  numroutes--;
  
  /* Allocate an array of routerec structs for routngi information	*/
  routes = (routerec *) malloc (numroutes * sizeof (routerec));
  
  rewind (f);
  
  fgets (buf, sizeof (buf), f);  /* strip out description line again	*/
  for (i = 0; i < numroutes; i++){
    if (fgets (buf, sizeof (buf), f) == NULL){
      /* Important, since an interface might have been removed since our counting,
	 causing us to parse bogus data */
      fputs ("Error reading /proc/net/route: iface count mismatch\n", stderr);
      fclose (f);
      exit (EXIT_FAILURE);
    }
    if ( strlen (buf) == sizeof(buf)-1 ){
      /* skip long lines */
      fputs ("Long (corrupt) line encountered, skipping.\n", stderr);
      while ((fgets (buf, sizeof (buf), f)))
	if (buf [strlen (buf) - 1] == '\n')
	  break;
      continue; /* continue with next regular line (or fail if EOF */
    }
    ptr = strtok (buf, "\t ");
    if (!ptr)
      continue;
    if (strlen (ptr) >= sizeof (iface))
      continue; /* would overflow if fed with bogus data in a chroot()ed environment */
    else
      strcpy (iface, ptr);
    ptr = strtok (NULL, "\t ");       /* hack avoiding fscanf */
    routes[i].addr.s_addr=(unsigned long)strtoul(ptr,NULL,16);
    for (i1 = 0; i1 < 6; i1++){
      ptr = strtok (NULL, "\t ");   /* ignore Gateway Flags RefCnt Use Metric */
    }
    if (!ptr) {
      fputs ("Error parsing /proc/net/route\n", stderr);
      continue;
    }
    routes[i].naddr=(unsigned long)strtoul(ptr,NULL,16);   /* Netmask */
    
    found = 0;
    for (i1 = 0; i1 < numinterfaces; i1++){
      if (strcmp (interfaces[i1].ifname, iface) == 0){
	routes[i].iface = &interfaces[i1];
	found = 1;
      }
      
    }
    
    if (!found){
      printf ("Couldn't find interface %s\n", iface);
      exit (EXIT_FAILURE);
    }
  }
  fclose (f);
}

/* end shameless theft of cheops code */

/* GLOBALS */
struct sockaddr_in addr1, addr2, addrtmp;
int src = 0, dst = 0, loc = 0; /* these will be set later. source, destination, and local ports */
int sock_type = SOCK_STREAM;  /* tells us we're going into udp mode :D */

/* debugging code */
int Debug(char *dbg){
  printf("%s\n", dbg);
}

/* sends out on a socket until ALL data is sent. :D */

int senda(int s, char *buf, int *len){
  int total = 0, rem = *len, n; /* vars for send */
  char msg[512];
  while(total < *len){
    Debug("Pass through send code");
    n = send(s, buf+total, rem, 0);
    if(n == -1) 
      break;
    total += n;
    rem -= n;
    sprintf(msg, "Total: %d\nRemaining: %d\nMessage: %s", total, rem, buf+total);
    Debug(msg);
  }

  *len = total;

  return (n==-1)? -1 : 0;
}

int get_out_socket(){
  int yes = 1;
  int sd2 = socket(AF_INET, sock_type, 0);

  if(src){
    if(setsockopt(sd2, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1){
      perror("setsockopt");
      exit(-1);
    }    
    if(bind(sd2, &addrtmp, sizeof(addrtmp)) != 0) {
      perror("forward bind port");
      exit(-1);
    }
  }

  return sd2;
}

/* s = source socket (listener)
   d = dest socket (receiver) */

int dopipe(int s){
  fd_set main;
  fd_set read_fds;
  struct sockaddr_in remote;
  int newfd = 0;
  char buf[256] = {0};
  int max, i, len;
  int d;

  max = s;

  for(;;){
    read_fds = main;
    if(select(max+1, &read_fds, NULL, NULL, NULL) < 0){
      perror("select");
      exit(1);
    }

    for(i = 0; i <= max; i++){
      if(FD_ISSET(i, &read_fds)){
	if(i == s){
	  /* maybe a connection?? */
	  if(newfd)
	    break;  /* only one connection */
	  len = sizeof(remote);
	  if((newfd = accept(s, (struct sockaddr *)&remote, &len)) < 0){
	    /* something went awry with the accept */
	    perror("accept");
	  } else {
	    printf("netpipe: connection accepted from %s\n", inet_ntoa(remote.sin_addr));
	    printf("making connection to remote host\n");
	    d = get_out_socket();
	    if(connect(d, (struct sockaddr *)&addr2, sizeof(struct sockaddr)) < 0){
	      perror("connecting");
	      exit(-1);
	    }
	    FD_SET(newfd, &main);
	    FD_SET(d, &main);
	    max = (newfd > max) ? newfd : max;
	    max = (d > max) ? d : max;
	  }
	} else {
	  /* i is now either d or newfd */
	  if((len = recv(i, buf, sizeof(buf), 0)) <= 0){
	    /* closed connection or otherwise */
	    if( !len ){
	      printf("Connection hung up\n");
	    }else{
	      perror("Socket error");
	    }
	    close(d);
	    close(newfd);
	    FD_CLR(d, &main);
	    FD_CLR(newfd, &main);
	    newfd = 0;
	    max = s;
	  } else {
	    if(i == newfd){
	      /* must send data now to d */
	      len = strlen(buf);
	      senda(d, buf, &len);
	    } else {
	      len = strlen(buf);
	      senda(newfd, buf, &len);
	    }
	  }
	}
      }
    }
  }
}

int main(int argc, char *argv[]){
  register int rr;
  int x;
  int yes = 1;  /* removal of time wasting */
  int need = 0; /* this is to make sure all needed options are passed */
  int sd1; /* these are the socket descriptors... sd1 = socket in, sd2 = socket out */
  int clientsd; /* needed for the accept() call */
  int optudp = 0;
  char *host; /* this is the ip or hostname of the machine to which it shall forward */

  if(argc < 5){
  needmore:
    printf("[%s] useage:\n"
	     "  %s  <-a redirect host> <-l port> [-s [port]] [-r [port]] [-p[T|U]]\n"
	     "    <-a redirect host>: This is the host that socket connections should be directed to.\n"
	     "    -l port: This is the port that should be listened on. Locally.\n"
	     "    -s port: This is the source port to make redirected connections from.\n"
	     "    -r port: This is the remote port to connect to. if NOT set, uses value from -l\n"
	     "    -p T|U:  Sets protocol to use.\n"
	     "    -h:      This help message.\n"
	     "Consult the manpages for more info. man %s at the command prompt.\n"
	     "Written by Aaron Conole\n", argv[0], argv[0], argv[0]);
    return -1;
  }

  while ((x = getopt (argc, argv, "h:a:l:s:r:p:M")) != EOF) {
    switch(x){
    case 'l':
      loc = atoi(optarg);
      need++;
      break;
    case 'a':
      host = (char *)malloc(sizeof(char) * strlen(optarg));
      strcpy(host, optarg);
      need++;
      break;
    case 's':
      src = atoi(optarg);
      break;
    case 'r':
      dst = atoi(optarg);
      break;
    case 'p':
      if((optarg[0] == 'U') || (optarg[0] == 'u'))
	optudp = 1;      
      break;
    case 'M':
      /* I need to do more work to get Multithreading properly working */
      break;
    default:
      printf("Unregocnized option : %c\n", x);
    case 'h':      /* fallthrough to a return */
      return 0;
    }
  }
  /* initialize routing information dewy */
  init_route_tables();

  bzero(&addr1, sizeof(addr1));
  bzero(&addr2, sizeof(addr2));
  bzero(&addrtmp, sizeof(addrtmp));

  addr1.sin_family = AF_INET;
  addr1.sin_port = htons(loc); /* local port...don't worry about this one */
  addr1.sin_addr.s_addr = INADDR_ANY;  /* listen for connects from anywhere (should be changed maybe) */

  addr2.sin_family = AF_INET;

  if( (addr2.sin_addr.s_addr = inet_addr(host)) < 0 ){
    perror("Error getting host lookup");
    exit(-1);
  }
 
  printf("Base configuration: forward connections to %s.\n"
	 "Listen locally on port %d\n", host, loc);

  if(src){
    printf("Retransmit from port %d\n", src);
    addrtmp.sin_family = AF_INET;
    addrtmp.sin_port = htons(src);
    /* get local IP address */
    addrtmp.sin_addr = getlocalip(addr2.sin_addr.s_addr);
  }
  
  if(dst){
    printf("Retransmit to port %d\n", dst);
    addr2.sin_port = htons(dst);
  }else{
    addr2.sin_port = htons(loc);
  }
  /* gentlemen..we have enough info..start your engines */
  
  /* first...create the 2 sockets... 1 listen, 1 for sending...later */
  if(optudp)
    sock_type = SOCK_DGRAM;

  sd1 = socket(AF_INET, sock_type, 0);

  /* address in use error is lossage...lets make that pesky thing
     <voice-of-guardian game="Ultima 8: The Black Gate">GO AWAY!</voice-of-guardian> */
  if(setsockopt(sd1, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1){
    perror("setsockopt");
    close(sd1);
    exit(-1);
  }

  if(bind(sd1, &addr1, sizeof(addr1)) != 0){
    perror("bind port");
    close(sd1);
    exit(-1);
  }

  if( listen(sd1, 0) != 0){
    perror("listen");
    close(sd1);
    return -1;
  }

  if(dopipe(sd1) < 0){
    close(sd1);
    return -1;
  }
  
  close(sd1);

  return 0;
}
